#!/usr/bin/env python 
#
#   ConVirt   -  Copyright (c) 2009 Convirture Corp.
#   ======
#
# ConVirt is a Virtualization management tool with a graphical user
# interface that allows for performing the standard set of VM operations
# (start, stop, pause, kill, shutdown, reboot, snapshot, etc...). It
# also attempts to simplify various aspects of VM lifecycle management.
#
#
# This software is subject to the GNU General Public License, Version 2 (GPLv2)
# and for details, please consult it at
#
#    http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt
#   
#   
# author : gizli
#    


from convirt.core.services.tasks import *
from convirt.core.services.core import Service, ServiceException
from convirt.core.services.execution_service import ThreadLimitException
from convirt.model.services import Task, TaskResult, TaskCalendar
from convirt.model.services import TaskInterval
from convirt.model import DBSession
from convirt.model.UpdateManager import UIUpdateManager

from convirt.model.notification import Notification
from convirt.model.auth import User
from datetime import datetime, timedelta
import time
import tg
from sqlalchemy.orm import eagerload
from convirt.core.utils.utils import to_unicode,to_str
import logging

logger = logging.getLogger("convirt.services.task_service")

class TaskManager(Service):

    IDLE = 0
    STARTED = 1
    FAILED = 2
    SUCCEEDED = 3

    def check_prior_execution(self, conn, t):
        #async submission requires us to check whether this task has been
        #executed before and whether that execution is still going on
        #We do not submit new executions of repeating tasks unless the prior
        #has completed
        res = conn.query(TaskResult).\
                filter(TaskResult.task_id == t.task_id).\
                filter(TaskResult.status == self.STARTED).\
                count()
        return (res == 0)


    def check_interval_tasks(self, conn):
        now = datetime.utcnow()
        try:
            for it in conn.query(TaskInterval).options(eagerload('task')).\
                      filter(TaskInterval.next_execution <= now):
                task = it.task
                if self.check_prior_execution(conn, task):
                    self.submit_async(task)
                    if it.interval and it.interval > 0:
                        it.next_execution = datetime.utcnow() + \
                                timedelta(minutes=it.interval)
                    else:
                        #if there is no interval, this is supposed to execute
                        #only once, so delete the interval entry
                        conn.delete(it)
                else:
                    logger.warning("The prior execution of task %s has not completed. Delaying the current execution..." % task.name)
        finally:
            conn.commit()

    def check_calendar_tasks(self, conn):
        now = list(time.gmtime())
        for cal in conn.query(TaskCalendar).options(eagerload('task')).\
                 filter((TaskCalendar.dow == 0) |\
                        TaskCalendar.dow.op('&')(1 << now[6]) > 0).\
                 filter((TaskCalendar.month == 0) |\
                        TaskCalendar.month.op('&')(1 << now[1]) > 0).\
                 filter((TaskCalendar.day == 0) |\
                        TaskCalendar.day.op('&')(1 << now[2]) > 0).\
                 filter((TaskCalendar.hour == 0) |\
                        TaskCalendar.hour.op('&')(1 << now[3]) > 0).\
                 filter((TaskCalendar.minute == 0) |\
                        TaskCalendar.minute.op('&')(1 << now[4]) > 0):
            task = cal.task
            if self.check_prior_execution(conn, task):
                self.submit_async(task)
            else:
                logger.warning("The prior execution of task %s has not completed. Delaying the current execution..." % task.name)
                #no update necessary

    def init(self):
        #clean-up previously started tasks
        #FIXME: This function needs to be improved to work in
        #multi CMS environment
        conn = self.get_database_conn()
        try:
            try:
                for incomplete_task in conn.query(TaskResult).\
                                       filter(TaskResult.status == self.STARTED):
                    incomplete_task.status = self.FAILED
                    incomplete_task.results = "Task Service was restarted before the task was completed."
                conn.commit()
            except Exception, e:
                logger.error("Error in cleaning up incomplete tasks: %s" % str(e))
                conn.rollback()
        finally:
            conn.close()


    def loop(self):
        #starter
        conn = self.get_database_conn()
        try:
            self.check_interval_tasks(conn)
            self.check_calendar_tasks(conn)
        finally:
            conn.close()

    def submit_schedule(self, t, execution_time):
        conn = self.get_database_conn()
        try:
            t.interval = [TaskInterval(interval=None,
                                       next_execution=execution_time),]
            conn.add(t)
            conn.commit()
        finally:
            conn.close()

    def submit_sync(self, t):
        conn = self.get_database_conn()
        try:
            conn.add(t)
            conn.commit()
            if not self.submit_task(t, 2):
                #in case of a failure we submit task asychronously
                t.interval = [TaskInterval(interval=None,
                                           next_execution=datetime.utcnow()),]
                conn.commit()
                raise ServiceException("All worker threads are busy. \
                                       Delaying task submission...")
        finally:
            conn.close()

    def submit_async(self, t):
        if not self.submit_task(t):
            self.task_fail(t.task_id, datetime.utcnow(),\
               "Task submission failed because all worker threads are \
                occupied. Please increase the number of threads.")

    def submit_task(self, t, max_sleep_time = \
                int(tg.config.get('services.task.max_sleep',20))):
        (this_service, target_service) = self.get_bridge("Execution Service")
        runtime_context = [this_service,]
        sleep_time = 0
        while sleep_time < max_sleep_time:
            try:
                target_service.new_work(t, ctx = runtime_context)
                return True
            except ThreadLimitException:
                logger.debug("Thread limit exception reached, trying again")
                # We try again
                time.sleep(0.5)
                sleep_time = sleep_time + 0.5
        return False

    def task_started(self, taskid, timestamp):
        conn = self.get_database_conn()
        try:
            res = TaskResult(taskid, timestamp, self.STARTED, None)
            conn.add(res)
            conn.commit()
            task=Task.get_task(taskid)            
            if len(task.interval) == 0:#to check if task is a repeating task or not
                UIUpdateManager().set_updated_tasks(taskid,task.user_name)
                
        finally:
            conn.close()

    def task_fail(self, taskid, ts, exception, user):
        conn = self.get_database_conn()
        try:
            results = to_str(exception)
            res = TaskResult(taskid, ts, self.FAILED, results)
            task=Task.get_task(taskid)
            u=User.by_user_name(user)
            email=u.email_address
            notification=Notification(taskid, task.name, ts,  results, user, email)
            conn.merge(res)
            conn.add(notification)
            conn.commit()
                        
            if len(task.interval) == 0:#to check if task is a repeating task or not
                UIUpdateManager().set_updated_tasks(taskid,task.user_name)
                
        finally:
            conn.close()

    def task_success(self, taskid, ts, results):
        conn = self.get_database_conn()
        try:
            res = TaskResult(taskid, ts, self.SUCCEEDED, results)
            conn.merge(res)
            conn.commit()
            task=Task.get_task(taskid)            
            if len(task.interval) == 0:#to check if task is a repeating task or not
                UIUpdateManager().set_updated_entities(task.get_entity_ids())
                UIUpdateManager().set_updated_tasks(taskid,task.user_name)
                
        finally:
            conn.close()
