#!/usr/bin/env python 
#
#   ConVirt   -  Copyright (c) 2009 Convirture Corp.
#   ======
#
# ConVirt is a Virtualization management tool with a graphical user
# interface that allows for performing the standard set of VM operations
# (start, stop, pause, kill, shutdown, reboot, snapshot, etc...). It
# also attempts to simplify various aspects of VM lifecycle management.
#
#
# This software is subject to the GNU General Public License, Version 2 (GPLv2)
# and for details, please consult it at
#
#    http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt
#   
#   
# author : gizli
#    


from convirt.model import DeclarativeBase, metadata, DBSession

from sqlalchemy import Column, Integer, String, Table
from sqlalchemy import ForeignKey, DateTime, PickleType, Boolean, Unicode
from sqlalchemy.orm import relation, backref
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.schema import Index
from datetime import datetime, timedelta
from convirt.model.Authorization import AuthorizationService
from convirt.model.auth import User
import transaction
from convirt.core.utils.utils import to_unicode,to_str
import logging

logger = logging.getLogger("convirt.services.model")

ImmutablePickleType = PickleType(mutable=False)

class PolymorphicSetter(DeclarativeMeta):
    def __new__(cls, name, bases, dictionary):
        if '__mapper_args__' in dictionary.keys():
            dictionary['__mapper_args__']['polymorphic_identity'] = to_unicode(name)
        else:
            dictionary['__mapper_args__'] = dict(polymorphic_identity=to_unicode(name))
        return DeclarativeMeta.__new__(cls, name, bases, dictionary)


##############################
## TASKS TABLES
##############################
#Processor interface
class Processor:
    def process_output(self, results):
        pass

class Task(DeclarativeBase):
    __tablename__ = 'tasks'
    __metaclass__ = PolymorphicSetter

    SHORT_DESC_KEY = "shortDesc"
    DESC_KEY = "desc"
    SHORT_DESC_PARAMS_KEY = "shortDescParams"
    DESC_PARAMS_KEY = "descParams"

    #FUTURE: When we introduce steps, they need to live in a separate table
    #since it does not make sense to introduce a relationship across the
    #rows of Tasks table. E.g. every step that belongs to a single task should
    #have the same calendar since they execute in the task context.
    task_id = Column(Integer, primary_key=True)
    task_type = Column(Unicode(50), default=to_unicode('Task'))
    name = Column(Unicode(256))
    entity_id = Column(Unicode(256))
    entity_name = Column(Unicode(50))
    context = Column(PickleType)
    params = Column(PickleType)
    kw_params = Column(PickleType)
    processors = Column(ImmutablePickleType)
    user_name = Column(Unicode(16))
    submitted_on = Column(DateTime)

    __mapper_args__ = {'polymorphic_on': task_type}

    def __init__(self, name, context, \
                 params, kw_params, processors, user_name):
        self.name = name
        self.context = context
        self.params = params
        self.kw_params = kw_params
        self.processors = processors
        self.user_name = user_name
        self.submitted_on = datetime.utcnow()
        descriptions = self.get_descriptions()
        if descriptions is not None:
            (short_desc, short_desc_params, \
             desc, desc_params) = descriptions
            self.set_short_desc(short_desc, short_desc_params)
            self.set_desc(desc, desc_params)

    def __repr__(self):
        return '<Task: name=%s>' % self.name

    def get_param(self, name):
        if self.kw_params is not None and \
           name in self.kw_params:
            return self.kw_params[name]

    def get_context_param(self, name):
        if self.context is not None and \
           name in self.context:
            return self.context[name]

    def set_short_desc(self, desc, desc_params):
        self.context[self.SHORT_DESC_KEY] = desc
        self.context[self.SHORT_DESC_PARAMS_KEY] = desc_params

    def set_desc(self, desc, desc_params):
        self.context[self.DESC_KEY] = desc
        self.context[self.DESC_PARAMS_KEY] = desc_params

    def get_short_desc(self):
        if self.context is not None and \
           self.SHORT_DESC_KEY in self.context:
            return (self.context[self.SHORT_DESC_KEY],\
                    self.context[self.SHORT_DESC_PARAMS_KEY])
        return None

    @classmethod
    def get_task(cls,task_id):        
        try:
            task=DBSession.query(cls).filter(cls.task_id==task_id).one()
            return task
        except Exception,ex:
            logger.error(to_str(ex).replace("'",""))
            traceback.print_exc()
            raise ex
    
    def get_desc(self):
        if self.context is not None and \
           self.DESC_KEY in self.context:
            return (self.context[self.DESC_KEY],\
                    self.context[self.DESC_PARAMS_KEY])
        return None

    def exec_task(self, context):
        raise Exception("exec_task needs to be implemented")

    def get_descriptions(self):
        raise Exception("get_descriptions needs to be implemented")

    def get_entity_ids(self):
        return self.entity_id

    def do_work(self, runtime_context, args = None, kw = None):
        task_manager = runtime_context.pop()
        curr_time = datetime.utcnow()
        args = self.params
        kw = self.kw_params

        if self.get_context_param('quiet') == True:
            task_manager = None

        if task_manager:
            task_manager.task_started(self.task_id, curr_time)

        try:
            try:
                if not args:
                    args = []
                if not kw:
                    kw = {}
                #build auth object
                auth = AuthorizationService()
                auth.user = User.by_user_name(self.user_name)
                raw_result = self.exec_task(auth, self.context, *args, **kw)
                #get processed results
                if raw_result is not None and self.processors is not None:
                    results = [raw_result,]
                    for p in self.processors:
                        if issubclass(p, Processor):
                            p().process_output(results)
                else:
                    results = raw_result
                #commit changes done by exec_task
                transaction.commit()
            except Exception, e:
                #rollback any changes made in thread session
                logger.exception(to_str(e))
                transaction.abort()
                if task_manager:
                    task_manager.task_fail(self.task_id, curr_time, e,self.user_name)
            else:
                if task_manager:
                    task_manager.task_success(self.task_id, curr_time, results)
        finally:
            DBSession.remove()

Index("task_id_eid", Task.task_id, Task.entity_id)


class TaskResult(DeclarativeBase):
    __tablename__ = 'task_results'

    task_id = Column(Integer, ForeignKey('tasks.task_id'), primary_key=True)
    timestamp = Column(DateTime, primary_key=True)
    endtime = Column(DateTime)
    status = Column(Integer)
    results = Column(PickleType)

    task = relation(Task, backref=backref('result'))

    def __init__(self, task_id, timestamp, status, results):
        self.task_id = task_id
        self.timestamp = timestamp
        self.endtime=datetime.utcnow()
        self.status = status
        self.results = results

    def __repr__(self):
        return '<Task with id %s and timestamp %s returned %s>' % \
                (self.task_id, self.timestamp, self.status)

Index("tr_tid", TaskResult.task_id)

class TaskInterval(DeclarativeBase):
    """ The TaskInterval entity defines regularly scheduled
    items which execute at intervals. It can also optionally define
    one time execution by passing an interval value of <= 0. Such interval
    objects will be deleted after the first execution.
    """
    __tablename__ = "task_intervals"

    cal_id = Column(Integer, primary_key=True)
    task_id = Column(Integer, ForeignKey('tasks.task_id'))
    interval = Column(Integer)
    next_execution = Column(DateTime)

    task = relation(Task, backref=backref('interval'))

    def __init__(self, interval, next_execution=None):
        self.interval = interval
        if next_execution:
            self.next_execution = next_execution
        elif interval is not None and interval > 0:
            self.next_execution = datetime.utcnow() \
                    + timedelta(minutes = interval)
        else:
            raise Exception("Either define a non-zero interval \
                            or a next execution timestamp")

class TaskCalendar(DeclarativeBase):
    __tablename__ = 'task_calendars'

    cal_id = Column(Integer, primary_key=True)
    task_id = Column(Integer, ForeignKey('tasks.task_id'))
    dow = Column(Integer)
    month = Column(Integer)
    day = Column(Integer)
    hour = Column(Integer)
    minute = Column(Integer)

    task = relation(Task, backref=backref('calendar'))

    def create_bits(self, numbers):
        k = 0
        for n in numbers:
            if n > 60 or n < 0:
                continue
            k = k | (1 << n)
        return k


    def __init__(self, dow, month, day, hour, minute):
        #assume all variables are lists
        self.dow = self.create_bits(dow)
        self.month = self.create_bits(month)
        self.day = self.create_bits(day)
        self.hour = self.create_bits(hour)
        self.minute = self.create_bits(minute)

#######################
## SERVICES TABLES
#######################
Dependencies_table = Table("dependencies", metadata,
  Column('dependent_id', Integer, ForeignKey('services.id'), primary_key=True),
  Column('parent_id', Integer, ForeignKey('services.id'), primary_key=True))

class ServiceItem(DeclarativeBase):
    __tablename__ = 'services'

    id = Column(Integer, primary_key=True)
    name = Column(Unicode(255))
    type = Column(ImmutablePickleType)
    executor = Column(ImmutablePickleType)
    enabled = Column(Boolean)

    dependents = relation("ServiceItem", secondary=Dependencies_table,\
                   primaryjoin=id == Dependencies_table.c.parent_id,\
                   secondaryjoin=Dependencies_table.c.dependent_id == id)

    parents = relation("ServiceItem", secondary=Dependencies_table,\
                   primaryjoin=id == Dependencies_table.c.dependent_id,\
                   secondaryjoin=Dependencies_table.c.parent_id == id)


    def __init__(self, name, type, executor, enabled):
        self.name = name
        self.type = type
        self.executor = executor
        self.enabled = enabled

    def __repr__(self):
        return '<Service: %s>' % self.name


