#!/usr/bin/env python
#
#   ConVirt   -  Copyright (c) 2008 Convirture Corp.
#   ======
#
# ConVirt is a Virtualization management tool with a graphical user
# interface that allows for performing the standard set of VM operations
# (start, stop, pause, kill, shutdown, reboot, snapshot, etc...). It
# also attempts to simplify various aspects of VM lifecycle management.
#
#
# This software is subject to the GNU General Public License, Version 2 (GPLv2)
# and for details, please consult it at:
#
#    http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt
# 
#
# author : Jd <jd_jedi@users.sourceforge.net>
#

# Manages shared storage definitions
# Each storage definition represents the shared storage as seen from the
# clients perspective.
# In future this would be enhanced to
# a. model the server view as well
# b. Separate SharedServer entity to which each of the storage definition
#    would belong

from datetime import datetime
from convirt.core.utils.utils import copyToRemote, getHexID, mkdir2
from convirt.core.utils.utils import to_unicode,to_str
import convirt.core.utils.utils
from convirt.core.utils.constants import *
constants = convirt.core.utils.constants
import os, tg, transaction
import pprint
from tg import session

from sqlalchemy import Table, ForeignKey, Column
from sqlalchemy.types import Integer, Unicode, String, Boolean, PickleType, Float, DateTime
from sqlalchemy.schema import UniqueConstraint,Index
from sqlalchemy.orm import relation, backref

#Import for Storage and Network classes
from convirt.model import DeclarativeBase, DBSession
from convirt.model.ManagedNode import ManagedNode
from convirt.model.Groups import ServerGroup
from convirt.model.SPRelations import ServerDefLink, SPDefLink, DCDefLink, StorageDisks
from convirt.model.VM import VMDisks, VMStorageLinks
from convirt.model.Entity import Entity, EntityRelation
from convirt.model.Sites import Site
from convirt.core.utils.utils import dynamic_map
from convirt.model.Authorization import AuthorizationService

from convirt.model.SyncDefinition import SyncDef

import logging
LOGGER = logging.getLogger("convirt.model")

class StorageDef(DeclarativeBase):
    __tablename__ = 'storage_definitions'

    #Columns
    id = Column(Unicode(50), primary_key=True)
    name = Column(Unicode(100), nullable=False)
    type = Column(Unicode(50), nullable=False)
    description = Column(Unicode(250))
    connection_props = Column(PickleType)
    creds_required = Column(Boolean)
    creds = Column(PickleType)
    stats = Column(PickleType)
    is_deleted = Column(Boolean)
    scope = Column(String(2))

    def __init__(self, id, name, type, description,
                connection_props, scope, creds_required = False, stats = {}, is_deleted = False, status = None):  

        self.id = id
        if self.id is None:
            self.id = getHexID()
        
        self.name = name
        self.type = type
        self.description = description

        self.connection_props = connection_props
        self.creds_required = creds_required
        self.creds = {}
        self.stats = stats # map containing TOTAL, USED and AVAIL
        self.total_size = 0
        self.is_deleted = is_deleted
        self.status = status
        self.scope = scope
        
    def sanitized_creds(self):
        if self.creds and self.creds.get("password"):
            new_creds = self.creds.copy()
            new_creds["password"] = None
            return new_creds
        return self.creds

    def __repr__(self):
        return to_str({"id":self.id,
                    "type":self.type,
                    "name":self.name,
                    "description":self.description,
                    "connection_props" : self.connection_props,
                    "creds_required":self.creds_required,
                    "creds" : self.sanitized_creds(),
                    "stats" : self.stats
                    })

    def get_connection_props(self):
        return self.connection_props

    def set_connection_props(self, cp):
        self.connetion_props = cp

    def get_creds(self):
        return self.creds

    def get_creds(self):
        return self.creds
    
    def set_creds(self, creds):
        self.creds = creds

    def set_stats(self, stats):
        self.stats = stats

    def get_stats(self):
        return self.stats

    def set_status(self, status):
        self.status = status

Index("strgedef_id", StorageDef.id)

class StorageManager:
    s_scripts_location = "/var/cache/convirt/storage"
    
    def __init__(self):
        self.storage_processors = { constants.NFS : self.nfs_processor,
                                    constants.iSCSI : self.iscsi_processor,
                                    constants.AOE : self.aoe_processor
                                    }

    def getType(self):
        return to_unicode(constants.STORAGE)

    # return the storage definitions for a given group    
    def getSiteDefListToAssociate(self, site_id, group_id, defType):
        sdArray=[]
        if site_id:
            dc_rs = DBSession.query(DCDefLink).filter_by(site_id=site_id, def_type=defType)
            if dc_rs:
                for row in dc_rs:
                    sp_def = DBSession.query(SPDefLink).filter_by(group_id=group_id, def_id=row.def_id, def_type=defType).first()
                    if not sp_def:
                        defn = DBSession.query(StorageDef).filter_by(id=row.def_id, scope=constants.SCOPE_DC).first()
                        if defn:
                            defn.status = row.status
                            sdArray.append(defn)
        return sdArray

    # return the storage definitions for a given group    
    def get_sds(self, site_id, group_id):
        if group_id:
            rs = DBSession.query(SPDefLink).filter_by(group_id=group_id)
        elif site_id:
            rs = DBSession.query(DCDefLink).filter_by(site_id=site_id)

        sdArray=[]
        for row in rs:
            row_nw = DBSession.query(StorageDef).filter_by(id=row.def_id).first()
            if row_nw:
                sdArray.append(row_nw)
        return sdArray

    # return list of definition ids in the group
    def get_sd_ids(self, site_id, group_id, defType, scope):
        ids_array=[]
        def_ids=None
        if scope == constants.SCOPE_SP or scope == constants.SCOPE_S:
            def_ids = DBSession.query(SPDefLink).filter_by(group_id=group_id, def_type = to_unicode(defType))
        elif scope == constants.SCOPE_DC:
            def_ids = DBSession.query(DCDefLink).filter_by(site_id=site_id, def_type = to_unicode(defType))
        
        if def_ids:
            for row in def_ids:
                ids_array.append(row.def_id)

        return ids_array

    #It return storage definitions
    def get_sd(self, sd_id, site_id, group_id, defType):
        status = None
        rsSD = DBSession.query(StorageDef).filter_by(id=sd_id).first()
        if rsSD:
            if rsSD.scope == constants.SCOPE_SP:
                group_defn = DBSession.query(SPDefLink).filter_by(group_id=group_id, def_type=to_unicode(defType), def_id=sd_id).first()
            elif rsSD.scope == constants.SCOPE_DC:
                group_defn = DBSession.query(DCDefLink).filter_by(site_id=site_id, def_type=to_unicode(defType), def_id=sd_id).first()
                
            if group_defn:
                status = group_defn.status    # getting definition status (IN_SYNC/ OUT_OF_SYNC) to display in grid on UI 

            rsSD.status = status
        return rsSD

    def get_defn(self, id):
        defn = DBSession.query(StorageDef).filter_by(id=id).first()
        return defn

    # execute the GET_DISK on the give node and return the results
    def get_sd_details(self, auth, sd, node, group, site, defType, def_manager):
        details = None
        groupId = None
        existing_def = None
        
        if group:
            groupId = group.id
        else:
            entity = auth.get_entity(node.id)
            if entity:
                groupId = entity.parents[0].entity_id

        sync_manager = SyncDef()
        try:
            #Get sync the definition and get disk details
            result_processor = self.storage_processors[sd.type]
            errs=[]
            op=constants.GET_DISKS
            update_status = False
            details = sync_manager.sync_node_defn(node,groupId, site.id, sd, to_unicode(defType), op, def_manager, update_status, errs, processor=result_processor)
        except Exception, ex:
            #mount point validation
            existing_def=self.isDefnExists(sd, node)
            #if there is a existing mount point then do not ATTACH it again.
            if existing_def:
                LOGGER.info("Definition already exists. This definition matches with the existing definition " + existing_def.name + " on the server " + node.hostname)
            
            #ATTACH here
            op = constants.ATTACH
            sync_manager.sync_node_defn(node,groupId, site.id, sd, defType, op, def_manager)
            
            #Get sync the definition and get disk details
            result_processor = self.storage_processors[sd.type]
            errs=[]
            op=constants.GET_DISKS
            update_status = False
            details = sync_manager.sync_node_defn(node,groupId, site.id, sd, defType, op, def_manager, update_status, errs, processor=result_processor)
            
            #DETACH here
            #if it is a existing mount point then do not DETACH it.
            if not existing_def:
                op = constants.DETACH
                sync_manager.sync_node_defn(node,groupId, site.id, sd, defType, op, def_manager)

        return details
        
    def parse_diskdetails(self,output, result):
        disk_details = []
        for i in output.splitlines():            
            d={}
            if not i:
                continue
            i = i.strip()
            if i.find("DISK_DETAILS") !=0:
                continue
            for j in i.split('|'):
                nameAndValue = j.lstrip().split('=')                
                d[nameAndValue[0]]= nameAndValue[1]
            del d['DISK_DETAILS']
            disk_details.append(d)
        return disk_details

    def parse_output(self,output, result):
        Lista = []
        for i in output.splitlines():
            d={}
            if not i:
                continue
            i = i.strip()
           
            if i.find("OUTPUT") !=0:
                continue
            for j in i.split('|'):
                nameAndValue = j.lstrip().split('=')
                d[nameAndValue[0]]= nameAndValue[1]
            del d['OUTPUT']
            Lista.append(d)
        return Lista

    def parse_summary(self,output, result):
        for i in output.splitlines():
            d={}
            if not i:
                continue
            i = i.strip()
           
            if i.find("SUMMARY") !=0:
                continue
            for j in i.split('|'):
                nameAndValue = j.lstrip().split('=')
                d[nameAndValue[0]]= nameAndValue[1]
            del d['SUMMARY']
            return d
        return None

    # storage processor
    def nfs_processor(self, op, output, result):
        print "nfs processor called with \n", output
        if op == "GET_DISKS_SUMMARY":
            result["SUMMARY"] = self.parse_summary(output,result)
        else:
            result["STORAGE_DETAILS"] = self.parse_output(output,result)
            result["DETAILS"] = self.parse_diskdetails(output,result)
            result["SUMMARY"] = self.parse_summary(output,result)
            
    def iscsi_processor(self, op, output, result):
        print "iscsi processor called with \n", output
        if op == "GET_DISKS_SUMMARY":
            result["SUMMARY"] = self.parse_summary(output,result)
        else:
           result["DETAILS"] = self.parse_output(output,result)
           result["SUMMARY"] = self.parse_summary(output,result)


    def aoe_processor(self, op, output, result):
        print "aoe processor called with \n", output
        if op == "GET_DISKS_SUMMARY":
            result["SUMMARY"] = self.parse_summary(output,result)
        else:
           result["DETAILS"] = self.parse_output(output,result)
           result["SUMMARY"] = self.parse_summary(output,result)

    def props_to_cmd_param(self, props):
        cp = ""
        if props:
            for p,v in props.iteritems():
                if v :
                    if cp:
                        cp += "|"
                    cp += "%s=%s" % (p,v)
            cp = "'%s'" % (cp, )
        return cp

    def prepare_scripts(self, dest_node, type, defType):
        s_src_scripts_location=tg.config.get("storage_script")
        s_src_scripts_location=os.path.abspath(s_src_scripts_location)

        LOGGER.info("Source script location= " + to_str(s_src_scripts_location))
        LOGGER.info("Destination script location= " + to_str(self.s_scripts_location))
        copyToRemote(s_src_scripts_location, dest_node, self.s_scripts_location)
        
        common_function_script_name = os.path.join(s_src_scripts_location,"storage_functions")
        LOGGER.info("Common source script location= " + to_str(common_function_script_name))
        
        dest_location = os.path.join(self.s_scripts_location,"scripts")
        LOGGER.info("Common destination script location= " + to_str(dest_location))
        
        copyToRemote(common_function_script_name, dest_node, dest_location, "storage_functions")

    def exec_script(self, node, group, defn, defType, op=constants.GET_DETAILS):
        type = defn.type

        self.prepare_scripts(node, type, defType)
        
        script_name = os.path.join(self.s_scripts_location,"scripts",
                                type, "storage.sh")
        
        log_dir = node.config.get(prop_log_dir)
        
        if log_dir is None or log_dir == '':
            log_dir = DEFAULT_LOG_DIR

        log_filename = os.path.join(log_dir, "storage/scripts",
                                    type, "storage_sh.log")
        
        # create the directory for log files
        mkdir2(node,os.path.dirname(log_filename))
        
        if type == constants.NFS:
            mount_point = defn.connection_props.get("mount_point")
            
            # create the directory for mount point for NFS mounting
            if mount_point.startswith("/") == True:
                mkdir2(node, mount_point)
        
        cp = self.props_to_cmd_param(defn.connection_props)
        
        creds_str = self.props_to_cmd_param(defn.creds)
        
        # help script find its location
        script_loc = os.path.join(self.s_scripts_location,"scripts")
        
        file_filter = "*.disk.xm"
        
        if type:
            script_args = " -t " + type
        if cp:
            script_args += " -c " + cp
        if creds_str:
            script_args += " -p " + creds_str
        if op:
            script_args += " -o " + op
        if script_loc:
            script_args += " -s " + script_loc
        if log_filename:
            script_args += " -l " + log_filename
        if file_filter:
            script_args += " -w " + file_filter

        
        cmd = script_name +  script_args
        LOGGER.info("Command= " + to_str(cmd))

        # execute the script
        output = "Success"
        exit_code = 0
        (output, exit_code) = node.node_proxy.exec_cmd(cmd)            

        LOGGER.info("Exit Code= " + to_str(exit_code))
        LOGGER.info("Output of script= " + to_str(output))
        return (exit_code, output)

    def CheckOp(self, op, errs):
        if not (op in [constants.GET_DISKS, constants.GET_DISKS_SUMMARY, constants.ATTACH, constants.DETACH]):
            errs.append("Invalid storage defn sync op " + op) 
            raise Exception("Invalid storage defn sync op " + op )
        return errs

    # This function checks whether there is a duplicate entry in the database for definition
    #Name the function
    def isDefnExists(self, defn, node):
        returnVal=None

        #get new def parameters
        con_props_new = defn.connection_props
        server_new = con_props_new.get("server")
        share_new = con_props_new.get("share")
        mount_point_new = con_props_new.get("mount_point")
        target_new = con_props_new.get("target")

        node_defns = DBSession.query(ServerDefLink).filter_by(server_id=node.id)
        for eachdefn in node_defns:
            sd = DBSession.query(StorageDef).filter_by(id=eachdefn.def_id).first()
            if sd:
                #get existing def parameters
                con_props = sd.connection_props
                server = con_props.get("server")
                share = con_props.get("share")
                local_server = node.hostname
                mount_point = con_props.get("mount_point")
                target = con_props.get("target")

                #checking duplicate here
                if defn.type == constants.NFS and server == server_new and share == share_new and mount_point == mount_point_new:
                    returnVal=sd
                elif defn.type == constants.iSCSI and server == server_new and target == target_new:
                    returnVal=sd
        return returnVal

    def getSyncMessage(self, op):
        messasge=None
        if op == constants.ATTACH:
            messasge="Mount operation is done successfully."
        elif op == constants.DETACH:
            messasge="Unmount operation is done successfully."
        return messasge
    
    def sync_node_defn(self, node, group_id, site_id, defn, defType, op, def_manager=None, update_status=True, errs=None, processor=None):
        pass

    def manage_storage_disk(self, storage_id, grid_manager):
        defn = self.get_defn(storage_id)
        if defn:
            objStats = defn.get_stats()
                    
            if objStats:
                objSummary = objStats.get("SUMMARY")
                objStats_details = objStats.get("DETAILS")
        
                if objStats_details:
                    for eachdetail in objStats_details:
                        actual_size = eachdetail.get("SIZE")
                        used_size = eachdetail.get("USED")
                        unique_path = eachdetail.get("uuid")
                        current_portal = eachdetail.get("CurrentPortal")
                        target = eachdetail.get("Target")
                        state = eachdetail.get("State")
                        lun = eachdetail.get("Lun")
                        
                        #decide here to add or update the disk record
                        storage_disk = DBSession.query(StorageDisks).filter_by(storage_id=storage_id, unique_path=unique_path).first()
                        if storage_disk:
                            #update record
                            if not used_size:
                                used_size=0
                            if used_size==0:
                                used_size=actual_size
                            storage_disk.size = float(used_size)
                            storage_disk.actual_size = float(actual_size)
                            storage_disk.state = state
                            storage_disk.lun = lun
                        else:
                            #add new record
                            storage_allocated = False
                            self.add_storage_disk(storage_id, actual_size, used_size, unique_path, current_portal, target, state, lun, storage_allocated, grid_manager)
                        
                        #delete record
                        #compare the stats object with database records and decide which disk is not in the stats
                        #object. Consider that disk is physically deleted. So remove the entry for database.
                        all_disks = DBSession.query(StorageDisks).filter_by(storage_id=storage_id)
                        if all_disks:
                            for each_disk in all_disks:
                                disk_found=False
                                #Assume that entry does not have any unique_path would not be a disk.
                                #Delete nfs container entry here.
                                if not str(each_disk.unique_path).strip():
                                    DBSession.delete(each_disk)
                                    continue
                                    
                                for each_stat in objStats_details:
                                    if str(each_stat.get("uuid")).strip() == str(each_disk.unique_path).strip():
                                        disk_found=True
                                
                                #This disk is deleted
                                if disk_found==False:
                                    vm_storage_link = DBSession.query(VMStorageLinks).filter_by(storage_disk_id=each_disk.id).first()
                                    if vm_storage_link:
                                        LOGGER.info("The storage disk (" + each_disk.disk_name + ") is in use so that it can not be deleted.")
                                    else:
                                        DBSession.delete(each_disk)
                        
    def is_storage_allocated(self, storage_id):
        returnVal=False
        if storage_id:
            disk = DBSession.query(StorageDisks).filter_by(storage_id=storage_id, storage_allocated=True).first()
            if disk:
                LOGGER.info("Storage (" + str(disk.disk_name) + ") is in use. It can not be deleted.")
                returnVal=True
        return returnVal
    
    def remove_storage_disk(self, storage_id):
        if storage_id:
            disks = DBSession.query(StorageDisks).filter_by(storage_id=storage_id)
            for eachdisk in disks:
                DBSession.delete(eachdisk)

    def calculate_disk_size(self, storage_disk_id, vm_disk_id, disk_size, op):
        vm_disk=None
        if not disk_size:
            disk_size=0
        
        if vm_disk_id:
            vm_disk = DBSession.query(VMDisks).filter_by(id=vm_disk_id).first()
            if vm_disk:
                vm_disk.size = self.convert_to_GB(disk_size)
        
        if storage_disk_id:
            storage_disk = DBSession.query(StorageDisks).filter_by(id=storage_disk_id).first()
            if storage_disk:
                storage_allocated = storage_disk.storage_allocated
                if op == "+":
                    #storage is unavailable now
                    if vm_disk:
                        if vm_disk.read_write == "w":
                            storage_allocated = True
                elif op == "-":
                    #Make it available only when the vm disk which is getting removed should be having mode read_write (w).
                    #storage is available for use now
                    if vm_disk:
                        if vm_disk.read_write == "w":
                            storage_allocated = False
                
                #storage availability is updated
                storage_disk.storage_allocated = storage_allocated
                
                #update storage_definitions table here for total used size and total available size 
                #for that storage id.
                self.update_size_in_storage_def(storage_disk.storage_id, storage_disk.unique_path, disk_size)

    def update_size_in_storage_def(self, storage_id, unique_path, used_size):
        defn = DBSession.query(StorageDef).filter_by(id=storage_id).first()
        if defn:
            stats = defn.stats
            if stats:
                return
            
                objStat={}
                objSummary={}
                objDetailsList=[]
                
                objSummary = stats["SUMMARY"]
                if not objSummary:
                    LOGGER.error("Error: SUMMARY object is not found. Can not update size in storage_definitions table.")
                    return
                    
                if float(objSummary.get("TOTAL"))>0:
                    total_size = self.convert_to_MB(objSummary.get("TOTAL"))
                if not total_size:
                    total_size=0
                    
                objDetailsList = stats["DETAILS"]
                if not objDetailsList:
                    LOGGER.error("Error: DETAILS object is not found. Can not update size in storage_definitions table.")
                    return
                
                if objDetailsList:
                    total_used_size=0
                    for each_disk in objDetailsList:
                        if each_disk.get("uuid"):
                            if each_disk.get("uuid") == unique_path:
                                if float(used_size)>0:
                                    each_disk["USED"] = self.convert_to_GB(used_size)
                                else:
                                    each_disk["USED"] = 0
                            if float(each_disk.get("USED"))>0:
                                total_used_size += self.convert_to_MB(each_disk.get("USED"))
                            objDetailsList.append(each_disk)
    
                    objStats["name"] = stats.get("name")
                    objStats["type"] = stats.get("type")
                    objStats["id"] = stats.get("id")
                    objStats["op"] = stats.get("op")
                    
                    #calculate available size here
                    available_size = (total_size - total_used_size)
                    objSummary["AVAILABLE"] = self.convert_to_GB(available_size)
                    
                    objStats["SUMMARY"] = objSummary
                    objStats["DETAILS"] = objDetailsList
                    
                defn.stats = objStats
        
    def convert_to_MB(self, size_in_GB):
        if not size_in_GB:
            size_in_GB=0

        size_in_GB = float(size_in_GB)
        if size_in_GB > 0:
            #since 1 GB is equal to 1024 MBs
            size = (size_in_GB * 1024)
        else:
            size = 0
        size = round(size, 2)
        return size
        
    def convert_to_GB(self, size_in_MB):
        if not size_in_MB:
            size_in_MB=0
            
        size_in_MB = float(size_in_MB)
        if size_in_MB > 0:
            #since 1024 MBs are equal to 1 GB
            size = (size_in_MB/1024)
        else:
            size = 0
        size = round(size, 2)
        return size

    def convert_to_GB_from_Bytes(self, size_in_Bytes):
        #since 1024 Bytes are equal to 1 KB
        #since 1024 KBs are equal to 1 MB
        #since 1024 MBs are equal to 1 GB
        GB_BYTES = 1024.0 * 1024.0 * 1024.0
        
        if not size_in_Bytes:
            size_in_Bytes=0
            
        size_in_Bytes = float(size_in_Bytes)
        if size_in_Bytes > 0:
            size = (size_in_Bytes/GB_BYTES)
        else:
            size = 0
        size = round(size, 2)
        return size

    def add_storage_disk(self, storage_id, actual_size, size, unique_path, current_portal, target, state, lun, storage_allocated, grid_manager):
        storage_disk_id=None
        if storage_id:
            mount_point=None
            file_system=None
            storage_type=None
            disk_name=None
            defn = self.get_defn(storage_id)
            if defn:
                disk_name = defn.name
                storage_type = defn.type
                objStats = defn.get_stats()
                objStorage_details = objStats.get("STORAGE_DETAILS")
                if objStorage_details:
                    mount_point = objStorage_details[0].get("MOUNT")
                    file_system = objStorage_details[0].get("FILESYSTEM")
    
            #add new record
            storage_disk = StorageDisks()
            storage_disk_id = getHexID()
            storage_disk.id = storage_disk_id
            storage_disk.storage_id = storage_id
            storage_disk.storage_type = storage_type
            storage_disk.disk_name = disk_name
            storage_disk.mount_point = mount_point
            storage_disk.file_system = file_system
            if not actual_size:
                actual_size=0
            if not size:
                size=0
            storage_disk.actual_size = float(actual_size)
            storage_disk.size = float(size)
            storage_disk.unique_path = unique_path
            storage_disk.current_portal = current_portal
            storage_disk.target = target
            storage_disk.state = state
            storage_disk.lun = lun
            storage_disk.storage_allocated = storage_allocated
            DBSession.add(storage_disk)
            #match to vm disk
            #if storage_disk_id:
            #    grid_manager.matching_disk_on_discover_storage(storage_disk_id)
        return storage_disk_id

    def RemoveScanResult(self):
        try:
            scan_result = session.get(constants.SCAN_RESULT)
            if scan_result:
                session[constants.SCAN_RESULT] = None
                session.save()
            return "{success:true, msg:''}"
        except Exception, ex:
            error_msg = to_str(ex).replace("'","")
            LOGGER.error(error_msg)
            return "{success:false, msg:'" + error_msg + "'}"

    def SaveScanResult(self, storage_id, grid_manager):
        try:
            scan_result = session.get(constants.SCAN_RESULT)
            if scan_result:
                defn = self.get_defn(storage_id)
                if defn:
                    defn.stats = scan_result
                    #add/update/remove entry of storage disk in storage_disks table
                    self.manage_storage_disk(defn.id, grid_manager)
                    LOGGER.info("Storage stats and storage disks are updated.")
                else:
                    LOGGER.info("Storage stats is not updated since storage is not found.")
            return "{success:true, msg:''}"
        except Exception, ex:
            error_msg = to_str(ex).replace("'","")
            LOGGER.error(error_msg)
            return "{success:false, msg:'" + error_msg + "'}"
